/****************************************************************************
 *
 * cuda-traffic-shared.c - Biham-Middleton-Levine traffic model for
 * the GPU, using shared memory and two kernels (one for LR vehicles
 * and one for TB vehicles).
 *
 * To cite this work, see the file CITATION
 *
 * Copyright (C) 2017, 2018, 2019 by Moreno Marzolla <moreno.marzolla(at)unibo.it>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/
#include "hpc.h"
#include <stdio.h>
#include <stdlib.h>     /* for rand(), srand()  */
#include <time.h>       /* for time()           */

#define BLKSIZE 32

typedef unsigned char cell_t;

__device__ int d_min(int a, int b)
{
    return (a < b ? a : b);
}

/* Possible values stored in a grid cell */
enum {
    EMPTY = 0,  /* empty cell                   */
    LR,         /* left-to-right moving vehicle */
    TB          /* top-to-bottom moving vehicle */
};

/* The following macro simplifies indexing of the grid arrays; it must
   be used inside a scope where the grid size |n| is defined. It is assumed
   that the grid is extended with ghost cells on each side; therefore,
   the actual size of the indexed grid is (n+2)*(n+2). Note: do not
   remove the parentheses. */
#define IDX(i,j) ((i)*((n)+2) + (j))

__global__ void copy_top_bottom(cell_t *grid, int n)
{
    const int i = 1 + threadIdx.y + blockIdx.y * blockDim.y;
    const int j = threadIdx.x + blockIdx.x * blockDim.x;
    if (i < 2 && j < n+2) { 
        grid[IDX(n+i, j)] = grid[IDX(i,     j)]; /* top to bottom halo */
        grid[IDX(i-1, j)] = grid[IDX(i-1+n, j)]; /* bottom to top halo */
    }    
}

__global__ void copy_left_right(cell_t *grid, int n)
{
    /* yes, i and j are swapped */
    const int i = threadIdx.x + blockIdx.x * blockDim.x;
    const int j = 1 + threadIdx.y + blockIdx.y * blockDim.y;
    if (i < n+2 && j < 2) {
        grid[IDX(i, n+j)] = grid[IDX(i, j    )]; /* left column to right halo */
        grid[IDX(i, j-1)] = grid[IDX(i, j-1+n)]; /* right column to left halo */
    }
}

__device__ cell_t lr_step(cell_t left, cell_t center, cell_t right)
{
    if ( left == LR && center == EMPTY )
        return LR;
    else {
        if ( center == LR && right == EMPTY )
            return EMPTY;
        else
            return center;
    }
}

__global__ void horizontal_step(cell_t *cur, cell_t *next, int n)
{
    __shared__ cell_t s_cur[BLKSIZE][BLKSIZE+2];

    /* "global" indexes */
    const int gi = 1 + threadIdx.y + blockIdx.y * blockDim.y;
    const int gj = 1 + threadIdx.x + blockIdx.x * blockDim.x;
    /* "local" indexes */
    const int li = threadIdx.y;
    const int lj = 1 + threadIdx.x;
    /* height and width of the subdomain handled by this thread
       block. Its maximum size is blockdim.x x blockDim.y, but could
       be less than that if the domain size is not a multiple of the
       block size. */
    const int height = d_min(blockDim.y, n+1-gi);
    const int width  = d_min(blockDim.x, n+1-gj);

    if ( gi < n+1 && gj < n+1 ) {        
        s_cur[li][lj] = cur[IDX(gi, gj)];
        if (lj == 1) { /* left-right */
            s_cur[li][0] = cur[IDX(gi, gj-1)];
            s_cur[li][1+width] = cur[IDX(gi, gj+width)];
        }
    }

    __syncthreads();

    if ( gi < n+1 && gj < n+1) {
        next[IDX(gi,gj)] = lr_step(s_cur[li][lj-1], s_cur[li][lj], s_cur[li][lj+1]);
    }
}

__device__ cell_t tb_step(cell_t top, cell_t center, cell_t bottom)
{
    if ( top == TB && center == EMPTY )
        return TB;
    else {
        if ( center == TB && bottom == EMPTY )
            return EMPTY;
        else
            return center;
    }
}

__global__ void vertical_step(cell_t *cur, cell_t *next, int n)
{
    __shared__ cell_t s_cur[BLKSIZE+2][BLKSIZE];

    /* "global" indexes */
    const int gi = 1 + threadIdx.y + blockIdx.y * blockDim.y;
    const int gj = 1 + threadIdx.x + blockIdx.x * blockDim.x;
    /* "local" indexes */
    const int li = 1 + threadIdx.y;
    const int lj = threadIdx.x;
    /* height and width of the subdomain handled by this thread
       block. Its maximum size is blockdim.x x blockDim.y, but could
       be less than that if the domain size is not a multiple of the
       block size. */
    const int height = d_min(blockDim.y, n+1-gi);
    const int width  = d_min(blockDim.x, n+1-gj);

    if ( gi < n+1 && gj < n+1 ) {        
        s_cur[li][lj] = cur[IDX(gi, gj)];
        if (li == 1) { /* top-bottom */
            s_cur[0][lj] = cur[IDX(gi-1,gj)];
            s_cur[1+height][lj] = cur[IDX(gi+height, gj)];
        }
    }

    __syncthreads();

    if (gi < n+1 && gj < n+1) {
        next[IDX(gi,gj)] = tb_step(s_cur[li-1][lj], s_cur[li][lj], s_cur[li+1][lj]);
    }
}

/* Returns 1 with probability p */
int rand01( float p )
{
    return ((rand() / (float)RAND_MAX) < p);
}

/* Initialize |grid| with vehicles with density |rho|. The direction of
   each vehicle is chosen with equal probability. */
void setup( cell_t* grid, int n, float rho )
{
    int i, j;
    for ( i=1; i<n+1; i++ ) {
        for ( j=1; j<n+1; j++) {
            if ( rand01( rho ) ) {
                grid[IDX(i,j)] = ( rand01(0.5) ? LR : TB );
            } else {
                grid[IDX(i,j)] = EMPTY;
            }
        }
    }
}

/* Dump |grid| as a PPM (Portable PixMap) image written to file
   |filename|. LR vehicles are shown as blue pixels, while TB vehicles
   are shown in red. Empty cells are white. */
void dump( const cell_t *grid, int n, const char* filename )
{
    int i, j;
    FILE *out = fopen( filename, "w" );
    if ( NULL == out ) {
        printf("Cannot create file %s\n", filename);
        abort();
    }
    fprintf(out, "P6\n");
    fprintf(out, "%d %d\n", n, n);
    fprintf(out, "255\n");
    for (i=1; i<n+1; i++) {
        for (j=1; j<n+1; j++) {
            switch( grid[IDX(i,j)] ) {
            case EMPTY:
                fprintf(out, "%c%c%c", 255, 255, 255);
                break;
            case TB:
                fprintf(out, "%c%c%c", 0, 0, 255);
                break;
            case LR:
                fprintf(out, "%c%c%c", 255, 0, 0);
                break;
            default:
                printf("Error: unknown cell state %u\n", grid[IDX(i,j)]);
                abort();
            }
        }
    }
    fclose(out);
}

#define BUFLEN 256

int main( int argc, char* argv[] )
{
    cell_t *cur;
    cell_t *d_cur, *d_next;
    double tstart, tend;
    const char *fname = "cuda-traffic-shared-%05d.ppm";
    char buf[BUFLEN];
    int s, N = 256, nsteps = 1024;
    float rho = 0.2;

    if ( argc > 4 ) {
        printf("Usage: %s [nsteps [rho [N]]]\n", argv[0]);
        return EXIT_FAILURE;
    }

    if ( argc > 1 ) {
        nsteps = atoi(argv[1]);
    }

    if ( argc > 2 ) {
        rho = atof(argv[2]);
    }

    if ( argc > 3 ) {
        N = atoi(argv[3]);
    }

    const size_t size = (N+2)*(N+2)*sizeof(cell_t);

    /* Grid size used for copying sides */
    dim3 cpyBlock(BLKSIZE);
    dim3 cpyGrid((N+2 + BLKSIZE-1)/BLKSIZE);
    /* Grid size used for the update step */
    dim3 stepBlock(BLKSIZE, BLKSIZE);
    dim3 stepGrid((N + BLKSIZE-1)/BLKSIZE, (N + BLKSIZE-1)/BLKSIZE);

    /* To add some non-determinism to the initialization phase,
       uncomment the following line */
    /* srand(time(NULL)); */

    /* Allocate grids on host */
    cur = (cell_t*)malloc(size);
    setup(cur, N, rho);

    /* Allocate grids on device */
    CudaSafeCall(cudaMalloc((void**)&d_cur, size));
    CudaSafeCall(cudaMalloc((void**)&d_next, size));

    /* Copy data to device */
    CudaSafeCall(cudaMemcpy(d_cur, cur, size, cudaMemcpyHostToDevice));

    tstart = hpc_gettime();
    for (s=0; s<nsteps; s++) {
#ifdef DUMPALL
        cudaMemcpy(cur, d_cur, size, cudaMemcpyDeviceToHost);
        snprintf(buf, BUFLEN, fname, s);
        dump(cur, N, buf);
#endif
        copy_left_right<<< cpyGrid, cpyBlock >>>(d_cur, N);
        horizontal_step<<< stepGrid, stepBlock >>>(d_cur, d_next, N);
        copy_top_bottom<<< cpyGrid, cpyBlock >>>(d_next, N);
        vertical_step<<< stepGrid, stepBlock >>>(d_next, d_cur, N);
    }
    cudaDeviceSynchronize();
    tend = hpc_gettime();
    const double elapsed = tend - tstart;
    const double nupdates = 1.0e-6 * N * N * nsteps;
    fprintf(stderr, "%s : %f Mupdates in %.4f seconds (%f Mupd/sec)\n", argv[0], nupdates, elapsed, nupdates/elapsed);

    CudaSafeCall(cudaMemcpy(cur, d_cur, size, cudaMemcpyDeviceToHost));
    snprintf(buf, BUFLEN, fname, s);
    dump(cur, N, buf);

    /* Free memory */
    free(cur);
    cudaFree(d_cur); cudaFree(d_next);
    return EXIT_SUCCESS;
}
