/****************************************************************************
 *
 * traffic.c - Biham-Middleton-Levine traffic model, scalar implementation
 *
 * To cite this work, see the file CITATION
 *
 * Copyright (C) 2017, 2018, 2019 by Moreno Marzolla <moreno.marzolla(at)unibo.it>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ---------------------------------------------------------------------------
 *
 * This program implements the Biham-Middleton-Levine traffic model
 * (see
 * https://en.wikipedia.org/wiki/Biham%E2%80%93Middleton%E2%80%93Levine_traffic_model
 * for details). THe BML traffic model is a simple three-state 2D
 * cellular automaton over a toroidal square lattice space.
 * Initially, each cell is either empty, or occupied by a
 * left-to-right (LR) or top-to-bottom (TB) moving vehicle. The model
 * evolves at discrete time steps. Each step is logically divided into
 * two phases: in the first phase only LR vehicles move, provided that
 * the destination cell is empty; in the second phase, only TB
 * vehicles move, again provided that the destination cell is
 * empty. The steady-state evolution of the model depends on the
 * initial conditions: if the density of vehicles is larger than a
 * critical threshold, a giant traffic jam eventually develops, and no
 * movement is possible. If the density is below the critical
 * threshold, the system self-organizes to allow a constant flow of
 * vehicles.
 *
 * Compile with:
 *
 * gcc -O3 -std=c99 -Wall -Wpedantic traffic.c -o traffic
 *
 * Run with:
 *
 * ./traffic [nsteps [rho [N]]]
 *
 * Example:
 *
 * ./traffic 1024 0.3
 * 
 * where nsteps is the number of simulation steps to execute, rho is
 * the density of vehicles (probability that a cell is occupied by a
 * vehicle), and N is the grid size.
 *
 * For each step, a file called traffic-XXXXX.ppm is generated, where NNNNN
 * is the step number (starts from 0). At the end of the simulation,
 * it is possible to display a movie with:
 *
 * animate -resize 512x512 traffic-*.ppm
 *
 * To make a movie with avconv:
 *
 * avconv -i "traffic-%05d.ppm" -vcodec mpeg4 -s 512x512 traffic.avi
 *
 ****************************************************************************/
#include "hpc.h"
#include <stdio.h>
#include <stdlib.h>     /* for rand(), srand()  */

typedef unsigned char cell_t;

/* Possible values stored in a grid cell */
enum {
    EMPTY = 0,  /* empty cell                   */
    LR,         /* left-to-right moving vehicle */
    TB          /* top-to-bottom moving vehicle */
};

/* The following macro simplifies indexing of the grid arrays; it must
   be used inside a scope where the grid size |n| is defined. Do not
   remove the parentheses. */
#define IDX(i,j) ((i)*(n) + (j))

/* Move all left-to-right vehicles that are not blocked */
void horizontal_step( cell_t *cur, cell_t *next, int n )
{
    int i, j;
    for (i=0; i<n; i++) {
        for (j=0; j<n; j++) {
            const cell_t left   = cur[IDX(i,(j-1+n)%n)];
            const cell_t center = cur[IDX(i,        j)];
            const cell_t right  = cur[IDX(i,  (j+1)%n)];
            cell_t *out = &next[IDX(i,j)];

            if ( left == LR && center == EMPTY ) {
                *out  = LR;
            } else {
                if ( center == LR && right == EMPTY ) {
                    *out = EMPTY;
                } else {
                    *out = center;
                }
            }
        }
    }
}

/* Move all top-to-bottom vehicles that are not blocked */
void vertical_step( cell_t *cur, cell_t *next, int n )
{
    int i, j;
    for (i=0; i<n; i++) {
        for (j=0; j<n; j++) {
            const cell_t top    = cur[IDX((i-1+n)%n,j)];
            const cell_t center = cur[IDX(i,        j)];
            const cell_t bottom = cur[IDX((i+1)%n,  j)];
            cell_t *out = &next[IDX(i,j)];

            if ( top == TB && center == EMPTY ) {
                *out = TB;
            } else {
                if ( center == TB && bottom == EMPTY) {
                    *out = EMPTY;
                } else {
                    *out = center;
                }
            }
        }
    }
}

/* Returns 1 with probability p */
int rand01( float p )
{
    return ((rand() / (float)RAND_MAX) < p);
}

/* Initialize |grid| with vehicles with density |rho|. The direction of
   each vehicle is chosen with equal probability. */
void setup( cell_t* grid, int n, float rho )
{
    int i, j;
    for ( i=0; i<n; i++ ) {
        for ( j=0; j<n; j++) {
            if ( rand01( rho ) ) {
                grid[IDX(i,j)] = ( rand01(0.5) ? LR : TB );
            } else {
                grid[IDX(i,j)] = EMPTY;
            }
        }
    }
}

/* Dump |grid| as a PPM (Portable PixMap) image written to file
   |filename|. LR vehicles are shown as blue pixels, while TB vehicles
   are shown in red. Empty cells are white. */
void dump( const cell_t *grid, int n, const char* filename )
{
    int i, j;
    FILE *out = fopen( filename, "w" );
    if ( NULL == out ) {
        printf("Cannot create file %s\n", filename);
        abort();
    }
    fprintf(out, "P6\n");
    fprintf(out, "%d %d\n", n, n);
    fprintf(out, "255\n");
    for (i=0; i<n; i++) {
        for (j=0; j<n; j++) {
            switch( grid[IDX(i,j)] ) {
            case EMPTY:
                fprintf(out, "%c%c%c", 255, 255, 255);
                break;
            case TB:
                fprintf(out, "%c%c%c", 0, 0, 255);
                break;
            case LR:
                fprintf(out, "%c%c%c", 255, 0, 0);
                break;
            default:
                printf("Error: unknown cell state %u\n", grid[IDX(i,j)]);
                abort();
            }
        }
    }
    fclose(out);
}

#define BUFLEN 256

int main( int argc, char* argv[] )
{
    cell_t *cur, *next;
    char buf[BUFLEN];
    const char *fname = "traffic-%05d.ppm";
    int s, N = 256, nsteps = 1024;
    float rho = 0.2;
    double tstart, tend;

    if ( argc > 4 ) {
        printf("Usage: %s [nsteps [rho [N]]]\n", argv[0]);
        return EXIT_FAILURE;
    }

    if ( argc > 1 ) {
        nsteps = atoi(argv[1]);
    }

    if ( argc > 2 ) {
        rho = atof(argv[2]);
    }

    if ( argc > 3 ) {
        N = atoi(argv[3]);
    }

    const size_t size = N*N*sizeof(cell_t);

    /* Allocate grids */
    cur = (cell_t*)malloc(size);
    next = (cell_t*)malloc(size);

    setup(cur, N, rho);
    tstart = hpc_gettime();
    for (s=0; s<nsteps; s++) {
#ifdef DUMPALL
        snprintf(buf, BUFLEN, fname, s);
        dump(cur, N, buf);
#endif
        horizontal_step(cur, next, N);
        vertical_step(next, cur, N);
    }
    tend = hpc_gettime();
    const double elapsed = tend - tstart;
    const double nupdates = 1.0e-6 * N * N * nsteps;
    fprintf(stderr, "%s : %f Mupdates in %.4f seconds (%f Mupd/sec)\n", argv[0], nupdates, elapsed, nupdates/elapsed);

    /* dump last state */
    snprintf(buf, BUFLEN, fname, s);
    dump(cur, N, buf);

    /* Free memory */
    free(cur);
    free(next);
    return EXIT_SUCCESS;
}
